"""Module that implements the Camera class."""
from datetime import datetime
from typing import Callable, List
from pyvivint.enums import CameraAttributes as Attributes
from pyvivint.devices import VivintDevice


class Camera(VivintDevice):
    """Represents a vivint camera device."""
    def __init__(self, data: dict, alarm_panel: 'pyvivint.devices.alarm_panel.AlarmPanel'):
        super().__init__(data, alarm_panel)
        self.__thumbnail_ready_callbacks: List[Callable] = list()

    @property
    def capture_clip_on_motion(self) -> bool:
        "Return True if capture clip on motion is active."
        return self.data[Attributes.CaptureClipOnMotion]

    @property
    def ip_address(self) -> str:
        "Camera's IP address."
        return self.data[Attributes.CameraIPAddress]

    @property
    def is_in_privacy_mode(self) -> bool:
        """Return True if privacy mode is active."""
        return self.data[Attributes.CameraPrivacy]

    @property
    def is_online(self) -> bool:
        """Return True if camera is online."""
        return self.data[Attributes.Online]

    @property
    def wireless_signal_strength(self) -> int:
        """Camera's wireless signal strength."""
        return self.data[Attributes.WirelessSignalStrenght]

    def add_thumbnail_ready_callback(self, callback: Callable) -> None:
        """Register a thumbnail_ready callback."""
        self.__thumbnail_ready_callbacks.append(callback)

    async def request_camera_thumbnail(self) -> None:
        """Request a new thumbnail for the camera."""
        await self.vivintskyapi.request_camera_thumbnail(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id
        )

    async def get_camera_thumbnail_url(self) -> str:
        """Returns the latest camera thumbnail URL."""
        camera_thumbnail_date = datetime.strptime(self.data[Attributes.CameraThumbnailDate], '%Y-%m-%dT%H:%M:%S.%fZ')
        thumbnail_timestamp = int(camera_thumbnail_date.timestamp() * 1000)

        return await self.vivintskyapi.get_camera_thumbnail_url(
            self.alarm_panel.id, self.alarm_panel.partition_id, self.id, thumbnail_timestamp
        )

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message addressed to this camera."""
        super().handle_pubnub_message(message)

        if message.get(Attributes.CameraThumbnailDate):
            self._fire_callbacks(self.__thumbnail_ready_callbacks)



"""The exceptions used by the pyvivint library."""


class VivintError(Exception):
    """General pyvivint exception occurred."""


class VivintSkyApiError(VivintError):
    """VivintSky API related exception occurred."""


class VivintSkyApiAuthenticationError(VivintSkyApiError):
    """VivintSky API authentication related error occurred."""


class VivintSkyApiExpiredTokenError(VivintSkyApiError):
    """VivintSky API token expired error occurred."""


class VivintSkyApiMissingTokenError(VivintSkyApiError):
    """VivintSky API missing id token error occurred."""



"""Module that implements the VivintSkyApi class."""
import aiohttp
import asyncio
import base64
import certifi
import datetime
import json
import logging
import os
import re
import ssl
import time

from http.cookies import SimpleCookie
from typing import Optional
from urllib.parse import urlencode, quote, unquote

from pyvivint.enums import ArmedStates, DoorLockAttributes
from pyvivint.exceptions import (
    VivintSkyApiError,
    VivintSkyApiAuthenticationError,
    VivintSkyApiExpiredTokenError,
    VivintSkyApiMissingTokenError
)

_LOGGER = logging.getLogger(__name__)

VIVINT_AUTH_ENDPOINT = 'https://id.vivint.com'
VIVINT_API_ENDPOINT = 'https://www.vivintsky.com'
USER_AGENT = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36'
BASE62_TABLE = (
    [chr(ord("a") + i) for i in range(26)]
    + [chr(ord("A") + i) for i in range(26)]
    + [chr(ord("0") + i) for i in range(10)]
)


class VivintSkyApi:
    """Class to communicate with the VivintSky API."""

    def __init__(
        self,
        username: str,
        password: str,
        loop: asyncio.events.AbstractEventLoop,
        client_session: Optional[aiohttp.ClientSession] = None,
    ):
        self.__username = username
        self.__password = password
        self.__loop = loop
        self.__client_session = client_session or self.__get_new_client_session()
        self.__has_custom_client_session = client_session is not None

        self.__openid_config: dict = None
        self.__openid_auth_data: dict = None
        self.__auth_user_data: dict = None

    @property
    def id_token(self) -> dict:
        """Return the current id_token."""
        if not self.__openid_auth_data:
            raise VivintSkyApiMissingTokenError('Missing openid auth data')

        return self.__openid_auth_data['id_token'][-1]

    async def connect(self) -> None:
        """Connect to VivintSky Cloud Service."""
        self.__openid_config = await self.__get_openid_config()
        self.__openid_auth_data = await self.__get_vivintsky_session(self.__username, self.__password)

    async def disconnect(self) -> None:
        """Disconnect from VivintSky Cloud Service."""
        if not self.__has_custom_client_session:
            await self.__client_session.close()

    def parse_id_token(self, id_token: str) -> dict:
        """
        Parse out the components of the ID token.
        """

        header_raw, payload_raw, data = [
            base64.urlsafe_b64decode(p + "=" * ((4 - (len(p) % 4)) % 4))
            for p in id_token.split(".")
        ]

        return {
            "header": json.loads(header_raw.decode()),
            "payload": json.loads(payload_raw.decode()),
            "data": base64.b64encode(data).decode(),
        }

    async def get_authuser_data(self) -> dict:
        """
        Get the authuser data

        Poll the Vivint authuser API endpoint resource to gather user-related data including enumeration of the systems
        that user has access to.
        """
        bearer_token = self.__get_bearer_token()
        resp = await self.__client_session.get(
            f'{VIVINT_API_ENDPOINT}/api/authuser',
            headers={
                'Authorization': f'Bearer {bearer_token}',
                'User-Agent': USER_AGENT,
            },
        )
        async with resp:
            return await resp.json(encoding='utf-8')

    async def refresh_token(self) -> None:
        """Refresh the openid token.

        Ping the token-delegate endpoint for a refresh token, geting the token-delegate
        endpoint from the OID configuration.
        """
        nonce = "".join([BASE62_TABLE[i % 62] for i in os.urandom(32)])
        state = "".join([BASE62_TABLE[i % 62] for i in os.urandom(32)])

        # When requesting a new token from the delegate, the following must be true:
        # - The original state and nonce must be in the Cookies header
        # - The new state and nonce must be in the query string
        #  > These don't appear to ever be used.
        # - The PF token supplied in the Cookies header must match that provided back
        #   in the Set-Cookie header from the original response to the auth ping
        #  > That is, not the "short" one, the "long" one.
        #
        # The response satisfies the following:
        # - The body is a JSON document with two keys: id_token, state
        # - The state value is as usual, "replay:%s"
        # - The new nonce to use is in the ID token returned.
        #  > Neither the new state or nonce need to be used anywhere.
        client_id = await self.__get_client_id()
        resp = await self.__client_session.get(
            self.__openid_config['token_delegate_endpoint'],
            params={
                'nonce': nonce,
                'state': f'replay:{state}',
                'response_type': 'id_token',
                'client_id': client_id,
                'redirect_uri': 'https://www.vivintsky.com/app/',
                'scope': 'openid email',
                'pfidpadapterid': 'vivintidp1',
            },
            headers={
                'Cookie': f'oidc_nonce={self.__openid_auth_data["nonce"][-1]}; oauth_state={self.__openid_auth_data["state"][-1]}; PF={self.__openid_auth_data["pf_token"]["long"][-1]}',  # noqa
                'User-Agent': USER_AGENT,
            },
        )
        async with resp:
            if resp.status == 200:
                resp_body = await resp.json()
                token_parts = self.parse_id_token(resp_body['id_token'])
                self.__openid_auth_data["id_token"].append(resp_body['id_token'])

                # It is the last valid value that we care about, because the new ones
                # generated aren't used for anything... oddly enough. So append the one
                # we used to get the new token to the end, so we keep using it.
                self.__openid_auth_data['state'] += [
                    unquote(resp_body['state']).split(":", 1)[1],
                    self.__openid_auth_data['state'][-2],
                ]
                self.__openid_auth_data['nonce'] += [
                    token_parts['payload']['nonce'],
                    self.__openid_auth_data['nonce'][-2],
                ]
                _LOGGER.info('vivint oauth token refreshed')
            else:
                # our session eventually expired. We need to re login
                # Attempt to re-login if there is a username and password
                if self.__username is not None and self.__password is not None:
                    # clear any pre-existing cookies for a fresh start
                    self.__client_session.cookie_jar.clear()
                    new_openid_auth_data = await self.__get_vivintsky_session(self.__username, self.__password)
                    self.__openid_auth_data['id_token'] += new_openid_auth_data['id_token']
                    self.__openid_auth_data['pf_token']['long'] += new_openid_auth_data['pf_token']['long']
                    self.__openid_auth_data['state'] += new_openid_auth_data['state']
                    self.__openid_auth_data['nonce'] += new_openid_auth_data['nonce']
                else:
                    raise VivintSkyApiAuthenticationError(
                        'Unable to refresh token with non-200 error, and no username/password available'
                    )

    async def get_system_data(self, panel_id: int) -> dict:
        """Gets the raw data for a system."""
        bearer_token = self.__get_bearer_token()
        resp = await self.__client_session.get(
            f'{VIVINT_API_ENDPOINT}/api/systems/{panel_id}',
            headers={
                'Authorization': f'Bearer {bearer_token}',
                'User-Agent': USER_AGENT,
            },
            params={'includerules': 'false'},
        )
        async with resp:
            return await resp.json(encoding='utf-8')

    async def set_alarm_state(self, panel_id: int, partition_id: int, state: bool) -> aiohttp.ClientResponse:
        bearer_token = self.__get_bearer_token()
        resp = await self.__client_session.put(
            f'{VIVINT_API_ENDPOINT}/api/{panel_id}/{partition_id}/armedstates',
            headers={
                'Authorization': f'Bearer {bearer_token}',
                'Content-Type': 'application/json;charset=UTF-8',
                'User-Agent': USER_AGENT,
            },
            data=json.dumps(
                {
                    'system': panel_id,
                    'partitionId': partition_id,
                    'armState': state,
                    'forceArm': False,
                }
            ).encode('utf-8'),
        )
        async with resp:
            if resp.status != 200:
                resp_body = await resp.text()
                _LOGGER.error(
                    f'failed to set state {ArmedStates.name(state)}. Code: {resp.status}, body: {resp_body},'
                    f'request url: {resp.request_info}'
                )
                raise VivintSkyApiError(
                    f'failed to set alarm status {ArmedStates.name(state)} for panel {self.id}'
                )

    async def set_lock_state(self, panel_id: int, partition_id: int, device_id: int, locked: bool) -> None:
        """Lock/Unlock door lock."""
        bearer_token = self.__get_bearer_token()
        resp = await self.__client_session.put(
            f'{VIVINT_API_ENDPOINT}/api/{panel_id}/{partition_id}/locks/{device_id}',
            headers={
                'Content-Type': 'application/json;charset=utf-8',
                'Authorization': f'Bearer {bearer_token}',
            },
            data=json.dumps({DoorLockAttributes.State: locked, DoorLockAttributes.Id: device_id}).encode('utf-8'),
        )
        async with resp:
            if resp.status != 200:
                _LOGGER.info(f'failed to set status locked: {locked} for lock: {device_id} @ {panel_id}:{partition_id}')
                raise VivintSkyApiError(f'failed to update lock status')

    async def request_camera_thumbnail(self, panel_id: int, partition_id: int, device_id: int) -> None:
        bearer_token = self.__get_bearer_token()
        resp = await self.__client_session.get(
            f'{VIVINT_API_ENDPOINT}/api/{panel_id}/{partition_id}/{device_id}/request-camera-thumbnail',
            headers={
                'Authorization': f'Bearer {bearer_token}',
                'User-Agent': USER_AGENT,
            },
        )
        async with resp:
            if resp.status < 200 or resp.status > 299:
                _LOGGER.info(f'failed to request thumbnail for camera id {self.id}. Error code: {resp.status}')

    async def get_camera_thumbnail_url(self, panel_id: int, partition_id: int, device_id: int,
                                       thumbnail_timestamp: datetime.datetime) -> str:
        bearer_token = self.__get_bearer_token()
        resp = await self.__client_session.get(
            f'{VIVINT_API_ENDPOINT}/api/{panel_id}/{partition_id}/{device_id}/camera-thumbnail',
            params={'time': thumbnail_timestamp},
            headers={
                'Authorization': f'Bearer {bearer_token}',
                'User-Agent': USER_AGENT,
            },
            allow_redirects=False,
        )

        async with resp:
            if resp.status != 302:
                _LOGGER.info(f'failed to request thumbnail for camera id {self.id}. Status code: {resp.status}')
                return

            return resp.headers.get('Location')

    def __get_new_client_session(self) -> aiohttp.ClientSession:
        """Create a new aiohttp.ClientSession object."""
        ssl_context = ssl.create_default_context(
            purpose=ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
        )
        connector = aiohttp.TCPConnector(
            loop=self.__loop, enable_cleanup_closed=True, ssl=ssl_context
        )

        return aiohttp.ClientSession(
            loop=self.__loop,
            connector=connector
        )

    async def __get_openid_config(self) -> dict:
        """Fetch the OpenID Connect configuration data from the VivintSky webservice."""
        resp = await self.__client_session.get(
            f'{VIVINT_API_ENDPOINT}/api/openid-configuration',
            headers={'User-Agent': USER_AGENT}
        )
        async with resp:
            return await resp.json(encoding='utf-8') if resp.status == 200 else None

    async def __get_client_id(self) -> str:
        """Fetch the client_id from the Vivint's authuser service."""
        # Fetch https://vivintsky.com/api/authuser and check the WWW-Authenticate header which has the client_id baked
        # in
        resp = await self.__client_session.get(
            f'{VIVINT_API_ENDPOINT}/api/authuser', headers={'User-Agent': USER_AGENT}
        )

        async with resp:
            if resp.status != 401:
                raise VivintSkyApiAuthenticationError(
                    f'Attempt to fetch authuser resulted in unexpected response code {resp.status}'
                )
            # noqa  Bearer scope="openid email",redirect_uri="https://www.vivintsky.com/api/oauth-redirect/[UUID]",realm="vivintsy",response_type="code",client_id="[UUID]"
            match = re.search(
                r'client_id="([0-9a-f]*)"', resp.headers['WWW-Authenticate']
            )
            if match is None:
                raise VivintSkyApiAuthenticationError(
                    'Unable to find client id within the response from https://vivintsky.com/api/authuser.'
                )

            client_id = match.group(1)

            return client_id

    async def __get_vivintsky_session(self, username: str, password: str) -> dict:
        """Login into the Vivint Sky platform with the given username and password.

        Returns an object that includes the appropriate OpenID components.
        """
        # As per the app.js, this is just random garbage
        nonce = "".join([BASE62_TABLE[i % 62] for i in os.urandom(32)])
        state = "".join([BASE62_TABLE[i % 62] for i in os.urandom(32)])

        client_id = await self.__get_client_id()

        oauth_data = urlencode(
            {
                'nonce': nonce,
                'state': f'replay:{state}',
                'response_type': 'id_token',
                'client_id': client_id,
                'redirect_uri': f'{VIVINT_API_ENDPOINT}/app/',
                'scope': 'openid email',
                'pfidpadapterid': 'vivintidp1',
            },
            quote_via=quote,
        )

        login_form_resp = await self.__client_session.get(
            f'{VIVINT_AUTH_ENDPOINT}/as/authorization.oauth2?{oauth_data}',
            headers={
                'Referer': f'{VIVINT_AUTH_ENDPOINT}/app/',
                'User-Agent': USER_AGENT,
            },
            allow_redirects=False,
        )
        async with login_form_resp:
            if login_form_resp.status == 302:
                # If we're being redirected it means we're already logged in and the PF token is in the session cokie.
                # We can get the other oauth data from the Location header
                location_hdr = login_form_resp.headers.get("Location", None)

                for cookie in self.__websession.cookie_jar:
                    if cookie.key == 'PF' and cookie['domain'] == 'id.vivint.com':
                        pf_token_long = cookie.value
                        break

                if pf_token_long is None:
                    raise VivintSkyApiAuthenticationError('Unable to get PF token from existing cookie')

            else:
                if login_form_resp.headers.get('Set-Cookie', None) is None:
                    raise VivintSkyApiAuthenticationError(
                        'Unable to get Set-Cookie/Cookie header from response'
                    )

                login_form_return_cookies = SimpleCookie(
                    login_form_resp.headers['Set-Cookie']
                )

                pf_token = login_form_return_cookies['PF'].value

                if pf_token is None:
                    raise VivintSkyApiAuthenticationError(
                        'Unable to get PF token from Set-Cookie/Cookie header'
                    )

                match = re.search(
                    r'"/as/([^/]*)/resume/as/authorization.ping"',
                    await login_form_resp.text(),
                )

                if match is None:
                    raise VivintSkyApiAuthenticationError('Unable to find api ID from login form HTML')

                api_id = match.group(1)

                login_url = (
                    f'{VIVINT_AUTH_ENDPOINT}/as/{api_id}/resume/as/authorization.ping'
                )

                # NOTABLE NOTE
                #
                # There's a weird thing with the Vivint API here on this call, where if there are nonce and state values
                # in the cookies and the query string, the cookies take precedence. This is odd since there's a weird
                # behaviour with their web JavaScript code that will provide both, with the code inheriting the cookies
                # from somewhere as well as generating its own values for the query string.
                #
                # The webapp will use the generated values (supplied in the query string) for further API calls, despite
                # being unnecessary given the bearer token is provided. But thankfully the correct values, being kept in
                # the tokens, are still used for all token refreshes to the delegate endpoint.
                login_resp = await self.__client_session.post(
                    login_url,
                    headers={
                        'Referrer': f'{VIVINT_AUTH_ENDPOINT}/app/',
                        'User-Agent': USER_AGENT,
                        'Accept': '*/*',
                        'Accept-Encoding': 'gzip, deflate',
                        'Content-Type': 'application/x-www-form-urlencoded',
                        'Cookie': f'oauth_state={state}; oidc_nonce={nonce}; PF={pf_token};',
                    },
                    data=urlencode(
                        {'pf.username': username, 'pf.pass': password}
                    ).encode('utf-8'),
                    allow_redirects=False,
                )
                async with login_resp:
                    login_resp_return_cookies = SimpleCookie(
                        login_resp.headers['Set-Cookie']
                    )

                    pf_token_long = login_resp_return_cookies['PF'].value

                    if pf_token_long is None:
                        raise VivintSkyApiAuthenticationError('Unable to get PF token from Set-Cookie header')

                    location_hdr = login_resp.headers.get('Location', None)

            if location_hdr is None:
                raise VivintSkyApiAuthenticationError(
                    'Unable to retrieve Location header from login response'
                )

            location_params = dict(
                [
                    kv.split('=', 1)
                    for kv in re.search(r'/#(.*)$', location_hdr)
                    .group(0)[2:]
                    .split('&')
                ]
            )

            if 'id_token' not in location_params:
                raise VivintSkyApiAuthenticationError(
                    'id_token not provided in Location header of login attempt'
                )

            if 'state' not in location_params:
                raise VivintSkyApiAuthenticationError(
                    'New state value not present in Location header of login attempt'
                )

            id_token_parts = self.parse_id_token(location_params['id_token'])

            return {
                'nonce': [nonce, id_token_parts['payload']['nonce']],
                'state': [state, unquote(location_params['state']).split(':', 1)[1]],
                'pf_token': {
                    'long': [pf_token_long]
                },
                'id_token': [location_params['id_token']]
            }

    def __get_bearer_token(self) -> dict:
        """
        Return a token suitable for inclusion into an Authorization header as
        a bearer token, refreshing the existing one if necessary.
        """
        id_token = self.__openid_auth_data["id_token"][-1]
        parsed_id_token = self.parse_id_token(id_token)
        if time.time() > parsed_id_token["payload"]["exp"]:
            raise VivintSkyApiExpiredTokenError()

        return id_token



"""Module that implements the Vivint class."""
import asyncio
import aiohttp
import json
import logging
import time

from typing import List, Optional

import pubnub
from pubnub.enums import PNReconnectionPolicy, PNHeartbeatNotificationOptions
from pubnub.pubnub_asyncio import PubNubAsyncio
from pubnub.pnconfiguration import PNConfiguration

import pyvivint.system
from pyvivint.enums import AuthUserAttributes, PubNumMessageAttributes as MessageAttributes
from pyvivint.pubnub import PN_CHANNEL, PN_SUBSCRIBE_KEY, VivintPubNubSubscribeListener
from pyvivint.system import System
from pyvivint.utils import add_async_job, first_or_none
from pyvivint.vivintskyapi import VivintSkyApi


_LOGGER = logging.getLogger(__name__)


class Vivint:
    """Class for interacting with VivintSky API using asyncio"""

    def __init__(
            self,
            username: str,
            password: str,
            loop: Optional[asyncio.events.AbstractEventLoop] = None,
            client_session: Optional[aiohttp.ClientSession] = None
    ):
        self.__loop: asyncio.events.AbstractEventLoop = loop or asyncio.get_event_loop()
        self.__pubnub: pubnub.pubnub_asyncio.PubNubAsyncio = None
        self.vivintskyapi = VivintSkyApi(username, password, self.__loop, client_session)
        self.systems: List[pyvivint.system.System] = []

    async def connect(
        self,
        load_devices: bool = False,
        subscribe_for_realtime_updates: bool = False,
        enable_token_auto_refresh: bool = False
    ) -> None:
        """Connects to vivintsky cloud service."""

        _LOGGER.debug('connecting to vivintsky')
        # initialize the vivintsky cloud session
        await self.vivintskyapi.connect()

        # load all systems, panels and devices
        if load_devices:
            _LOGGER.debug('loading devices')
            await self.refresh()

        # set up loop to periodically refresh the token
        if enable_token_auto_refresh:
            _LOGGER.debug('setting up token auto refresh loop')
            self.setup_token_refresh_handler()

        # subscript to pubnub for realtime updates
        if subscribe_for_realtime_updates:
            _LOGGER.debug('subscribing to pubnub for realtime updates')
            await self.subscripbe_for_realtime_updates()

    async def disconnect(self) -> None:
        _LOGGER.debug('disconnecting from vivintsky')
        await self.vivintskyapi.disconnect()

    async def refresh(self) -> None:
        # make a call to vivint's userauth endpoint to get a list of all the system_accounts (locations) & panels
        authuser_data = await self.vivintskyapi.get_authuser_data()

        # for each system_account, make another call to load all the devices
        for system_data in authuser_data['u']['system']:
            # is this an existing account_system?
            system = first_or_none(self.systems, lambda system: system.id == system_data['panid'])
            if system:
                await system.refresh()
            else:
                full_system_data = await self.vivintskyapi.get_system_data(system_data['panid'])
                self.systems.append(System(full_system_data, self.vivintskyapi))

        _LOGGER.debug(f'loaded {len(self.systems)} system(s)')

    def setup_token_refresh_handler(self) -> None:
        """Set up the token refresh handler."""
        add_async_job(self.__token_refresh_handler)

    async def __token_refresh_handler(self) -> None:
        """Periodically refreshes the idtoken."""
        while True:
            id_token = self.vivintskyapi.parse_id_token(self.vivintskyapi.id_token)
            if id_token['payload']['exp'] - time.time() < 60:
                await self.vivintskyapi.refresh_token()

            await asyncio.sleep(1)

    async def subscripbe_for_realtime_updates(self) -> None:
        """Subscribes to PubNub for realtime updates."""
        # make a call to vivint's userauth endpoint
        authuser_data = await self.vivintskyapi.get_authuser_data()

        pubnub.set_stream_logger("pubnub", logging.INFO)

        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = PN_SUBSCRIBE_KEY
        pnconfig.ssl = True
        pnconfig.reconnect_policy = PNReconnectionPolicy.LINEAR
        pnconfig.heartbeat_notification_options = PNHeartbeatNotificationOptions.ALL

        self.__pubnub = PubNubAsyncio(pnconfig)
        self.__pubnub.add_listener(
            VivintPubNubSubscribeListener(self.handle_pubnub_message)
        )

        pn_channel = f'{PN_CHANNEL}#{authuser_data[AuthUserAttributes.Users][AuthUserAttributes.UsersAttributes.MessageBroadcastChannel]}'  # noqa
        self.__pubnub.subscribe().channels(pn_channel).with_presence().execute()

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""
        _LOGGER.info(f"message: {json.dumps(message)}")

        panel_id = message.get(MessageAttributes.PanelId)
        if not panel_id:
            _LOGGER.info('ignoring pubnub message. No panel_id specified')
            return

        system = first_or_none(self.systems, lambda system: system.id == panel_id)
        if not system:
            _LOGGER.info(f'no system found with id {panel_id}')
            return

        system.handle_pubnub_message(message)



"""Module that implements the System class."""
import logging

from typing import List
from pyvivint.devices.alarm_panel import AlarmPanel
from pyvivint.entity import Entity
from pyvivint.enums import PubNumMessageAttributes as MessageAttributes
from pyvivint.utils import first_or_none
import pyvivint.vivintskyapi


_LOGGER = logging.getLogger(__name__)


class System(Entity):
    """Describe a vivint system."""

    def __init__(self, data: dict, vivintskyapi: pyvivint.vivintskyapi.VivintSkyApi):
        super().__init__(data)
        self.vivintskyapi = vivintskyapi
        self.alarm_panels: List[AlarmPanel] = [
            AlarmPanel(panel_data, self)
            for panel_data in self.data['system']['par']
        ]

    @property
    def id(self) -> str:
        """System's id"""
        return self.data['system']['panid']

    async def refresh(self) -> None:
        """Reloads system's data from VivintSky API."""
        system_data = await self.vivintskyapi.get_system_data(self.id)

        for panel_data in system_data['system']['par']:
            alarm_panel = first_or_none(
                self.alarm_panels,
                lambda panel: panel.id == panel_data['panid'] and panel.partition_id == panel_data['parid']
            )
            if alarm_panel:
                alarm_panel.refresh(panel_data)
            else:
                self.alarm_panels.append(AlarmPanel(panel_data, self))

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""

        if message[MessageAttributes.Type] == 'account_system':
            # this is a system message
            operation = message.get(MessageAttributes.Operation)
            data = message.get(MessageAttributes.Data)

            if data and operation == 'u':
                self.update_data(data)

        elif message[MessageAttributes.Type] == 'account_partition':
            # this is a message for one of the devices attached to this system
            partition_id = message.get(MessageAttributes.PartitionId)
            if not partition_id:
                _LOGGER.debug(f'ignoring account_partition message. No partition_id specified for system {self.id}')
                return

            alarm_panel = first_or_none(self.alarm_panels,
                                        lambda panel: panel.id == self.id and panel.partition_id == partition_id)

            if not alarm_panel:
                _LOGGER.debug(f'no alarm panel found for system {self.id}, partition {partition_id}')
                return

            alarm_panel.handle_pubnub_message(message)



"""Module that defines varios enums."""


class Enum:

    __names = {}

    @classmethod
    def name(cls, state):
        if not cls.__names:
            cls.__names = {
                value: key
                for key, value in cls.__dict__.items()
                if isinstance(value, int)
            }

        return cls.__names[state]


class VivintDeviceAttributes(Enum):
    """Devine generic vivint device attributes for humans."""
    DeviceType = 't'
    Id = '_id'
    Name = 'n'


class AlarmPanelAttributes(VivintDeviceAttributes):
    """Define alarm panel's attributes for humans."""
    Devices = 'd'
    PanelId = 'panid'
    PartitionId = 'parid'
    State = 's'


class ArmedStates(Enum):
    """Define all possible alarm states."""

    Disarmed = 0
    ArmingAwayInExitDelay = 1
    ArmingStayInExitDelay = 2
    ArmedStay = 3
    ArmedAway = 4
    ArmedStayInEntryDelay = 5
    ArmedAwayInEntryDelay = 6
    Alarm = 7
    AlarmFire = 8
    Disabled = 11
    WalkTest = 12


class CameraAttributes(VivintDeviceAttributes):
    """Define camera's attributes for humans."""

    CameraIPAddress = 'caip'
    CameraPrivacy = 'cpri'
    CameraThumbnailDate = 'ctd'
    CaptureClipOnMotion = 'ccom'
    Online = 'ol'
    WirelessSignalStrenght = 'wiss'


class DoorLockAttributes(VivintDeviceAttributes):
    """Define door lock's attributes for humans."""

    BatteryLevel = 'bl'
    Id = '_id'
    LowBattery = 'lb'
    OperationCount = 'opc'
    OperationFaultCode = 'opfc'
    NodeOnline = 'nonl'
    State = 's'


class DeviceTypes(Enum):
    """Vivint's supported device types."""
    Camera = 'camera_device'
    DoorLock = 'door_lock_device'
    LightModule = 'multilevel_switch_device'
    Thermostat = 'thermostat_device'
    TouchPanel = 'primary_touch_link_device'
    WirelessSensor = 'wireless_sensor'


class WirelessSensorAttributes(VivintDeviceAttributes):
    """Wireless Sensor's attributes for humans."""
    BatteryLevel = 'bl'
    EquipmentCode = 'ec'
    IdBypassed = 'b'
    LowBattery = 'lb'
    SerialNumber = 'ser32'
    State = 's'


class AuthUserAttributes(Enum):
    """AuthUser attributes."""
    IdToken = 'id_token'
    IsReadOnly = 'is_read_only'
    KeepSignedIn = 'keep_signed_in'
    RelayServer = 'rs'
    Users = 'u'

    class UsersAttributes(Enum):
        DocumentSequence = 'DocumentSequence'
        Email = 'e'
        Ghome = 'ghome'
        GroupIds = 'grpid'
        Id = '_id'
        MessageBroadcastChannel = 'mbc'
        Name = 'n'
        PingId = 'pngid'
        RestrictedSystem = 'rsystem'
        SmarthomeSystem = 'smarthomesystem'
        Settings = 'stg'
        System = 'system'
        Timestamp = 'ts'


class PubNumMessageAttributes(Enum):
    Data = 'da'
    Devices = 'd'
    Operation = 'op'
    PanelId = 'panid'
    PartitionId = 'parid'
    Type = 't'



"""This package contains the various devices attached to a vivint system."""
from typing import Callable

from pyvivint.enums import DeviceTypes, VivintDeviceAttributes as Attributes
from pyvivint.entity import Entity
import pyvivint.vivintskyapi


def get_device_class(device_type: str) -> Callable:
    """Maps a device_type string to the class that implements that device."""
    from pyvivint.devices import UnknownDevice
    from pyvivint.devices.camera import Camera
    from pyvivint.devices.door_lock import DoorLock
    from pyvivint.devices.wireless_sensor import WirelessSensor

    mapping = {
        DeviceTypes.Camera: Camera,
        DeviceTypes.DoorLock: DoorLock,
        DeviceTypes.WirelessSensor: WirelessSensor
    }

    return mapping.get(device_type, UnknownDevice)


class VivintDevice(Entity):
    """Class to implement a generic vivint device."""
    def __init__(self, data: dict, alarm_panel: 'pyvivint.devices.alarm_panel.AlarmPanel' = None):
        super().__init__(data)
        self.alarm_panel = alarm_panel

    def __repr__(self):
        """Custom repr method"""
        return f'<{self.__class__.__name__} {self.id}, {self.name}>'

    @property
    def id(self) -> int:
        """Device's id."""
        return self.data[Attributes.Id]

    @property
    def name(self) -> str:
        """Device's name."""
        return self.data[Attributes.Name]

    @property
    def vivintskyapi(self) -> pyvivint.vivintskyapi.VivintSkyApi:
        """Instance of VivintSkyApi."""
        assert self.alarm_panel, """no alarm panel set for this device"""
        return self.alarm_panel.system.vivintskyapi


class UnknownDevice(VivintDevice):
    """Describe an unknown/unsupported vivint device."""

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}|{self.data[Attributes.DeviceType]} {self.id}>'



"""Module that implements the AlarmPanel class."""
import logging
from typing import List, Set

from pyvivint.devices import VivintDevice, UnknownDevice, get_device_class
from pyvivint.enums import (
    ArmedStates,
    AlarmPanelAttributes as Attributes,
    PubNumMessageAttributes as MessageAttributes,
)
from pyvivint.utils import first_or_none
import pyvivint.vivintskyapi


_LOGGER = logging.getLogger(__name__)


class AlarmPanel(VivintDevice):
    """Describe a Vivint alarm panel."""

    def __init__(self, data: dict, system):
        super().__init__(data)
        self.system = system

        # initialize devices
        self.devices: List[VivintDevice] = [
            get_device_class(device_data[Attributes.DeviceType])(device_data, self)
            for device_data in self.data[Attributes.Devices]
        ]

    @property
    def vivintskyapi(self) -> pyvivint.vivintskyapi.VivintSkyApi:
        """Instance of VivitSkyApi."""
        return self.system.vivintskyapi

    @property
    def id(self) -> int:
        """Panel's id."""
        return self.data[Attributes.PanelId]

    @property
    def partition_id(self) -> int:
        """Panel's partition id."""
        return self.data[Attributes.PartitionId]

    @property
    def is_disarmed(self) -> bool:
        """Return True if alarm is disarmed."""
        return self.get_armed_state() == ArmedStates.Disarmed

    @property
    def is_armed_away(self) -> bool:
        """Return True if alarm is in armed away state."""
        return self.get_armed_state() == ArmedStates.ArmedAway

    def is_armed_stay(self) -> bool:
        """Return True if alarm is in armed stay state."""
        return self.get_armed_state() == ArmedStates.ArmedStay

    def get_armed_state(self):
        """Return the panel's arm state."""
        return self.data[Attributes.State]

    async def set_armed_state(self, state: int) -> None:
        """Set the armed state for a panel."""
        _LOGGER.debug(f"setting state: {ArmedStates.name(state)} for panel: {self.id}")
        await self.vivintskyapi.set_alarm_state(self.id, self.partition_id, state)

    async def disarm(self) -> None:
        """Disarm the alarm."""
        await self.set_armed_state(ArmedStates.Disarmed)

    async def arm_stay(self) -> None:
        """Set the alarm to armed stay."""
        await self.set_armed_state(ArmedStates.ArmedStay)

    async def arm_away(self) -> None:
        """Set the alarm to armed away."""
        await self.set_armed_state(ArmedStates.ArmedAway)

    def get_devices(self, device_types: Set[int] = None, include_unknown_devices: bool = False) -> List[VivintDevice]:
        """Get a list of associated devices."""
        devices = self.devices

        if device_types:
            devices = [
                device for device in self.devices if device.data[Attributes.DeviceType] in device_types
            ]

        if not include_unknown_devices:
            devices = [
                device for device in devices if not isinstance(device, UnknownDevice)
            ]

        return devices

    def refresh(self, data: dict) -> None:
        """Refreshes the alarm panel."""
        self.update_data(data, override=True)

        # update all associated devices
        for device_data in self.data[Attributes.Devices]:
            device = first_or_none(
                    self.devices, lambda device: device.id == device_data[Attributes.Id]
            )
            if device:
                device.update_data(device_data, override=True)
            else:
                device = get_device_class(device_data[Attributes.DeviceType])(device_data, self)
                self.devices.append(device)

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""
        data = message.get(MessageAttributes.Data)
        if not data:
            _LOGGER.debug(
                f"ignoring account_partition message for panel {self.id}, partition {self.partition_id} - "
                "no data provided"
            )
            return

        if not data.get(MessageAttributes.Devices):
            # this message is for the panel itself
            self.update_data(data)
        else:
            # this is a message for one (or more) of the attached devices
            devices_data = data[MessageAttributes.Devices]

            for device_data in devices_data:
                device_id = device_data.get("_id")
                if not device_id:
                    _LOGGER.debug("no device id")
                    continue

                device = first_or_none(
                    self.devices, lambda device: device.id == device_id
                )
                if not device:
                    _LOGGER.debug(
                        f"ignoring message for device {device_id}. Device not found"
                    )
                    continue

                device.handle_pubnub_message(device_data)

                # for the sake of consistency, we need to also update the panel's raw data
                raw_device_data = first_or_none(
                    self.data[Attributes.Devices],
                    lambda raw_device_data: raw_device_data["_id"]
                    == device_data["_id"],
                )
                raw_device_data.update(device_data)



"""Module that implements the WirelessSensor class."""
import logging

from pyvivint.devices import VivintDevice
from pyvivint.enums import WirelessSensorAttributes as Attributes


_LOGGER = logging.getLogger(__name__)


class WirelessSensor(VivintDevice):
    """Represents a Vivint wireless sensor device."""

    DEVICE_TYPE_TO_EQUIPMENT_CODE = {
        'COSensor': ['1026', '1266', '692'],
        'EntrySensor': ['1251', '1252', '862', '863', '655'],
        'FloodSensor': ['1128', '1264', '556'],
        'GarageDoor': ['1061', '2831'],
        'GlassBreakSensor': ['1248', '475', '864'],
        'HeatSensor': ['708'],
        'KeyFob': ['866', '577', '1250'],
        'Keypad': ['867'],
        'MotionSensor': ['1249', '609'],
        'PanicSensor': ['1253', '561', '868'],
        'Sensor': [
            '2830',
            '2832',
            '873',
            '941',
            '1208',
            '2081',
            '1144',
            '869',
            '1063',
            '1269',
            '0',
        ],
        'SmokeSensor': ['1058', '1066', '1267', '616'],
    }

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}|{self.device_class} {self.id}, {self.name}>'

    @property
    def battery_level(self) -> int:
        """Sensor's battery level."""
        return self.data[Attributes.BatteryLevel]

    @property
    def device_class(self):
        equipment_code = self.data.get(Attributes.EquipmentCode)
        if not equipment_code:
            _LOGGER.debug("device has no equipment code")
            return

        for device_type, codes in self.DEVICE_TYPE_TO_EQUIPMENT_CODE.items():
            if str(equipment_code) in codes:
                return device_type

    @property
    def is_bypassed(self) -> bool:
        """Return True if the sensor is bypassed."""
        return self.data[Attributes.IdBypassed]

    @property
    def is_on(self) -> bool:
        """Return True if the sensor's state is on."""
        return self.data[Attributes.State]

    @property
    def low_battery(self) -> bool:
        """Return true if battery's level is low."""
        return self.data[Attributes.LowBattery]

    @property
    def serial_number(self) -> str:
        """Return sensor' serial number."""
        return self.data[Attributes.SerialNumber]



#!/usr/bin/env python
"""pyvivint setup script."""

from setuptools import setup, find_packages

install_requires = [
    'aiohttp>=3.6',
    'certifi>=2019.9.11'
]

setup(
    name='pyvivint',
    version='0.1.6',
    description='Python module to interact with VivintSky API.',
    long_description=open('README.md').read(),
    author='Ovidiu Stateina',
    author_email='ovidiurs@gmail.com',
    license='MIT',
    install_requires=install_requires,
    packages=find_packages(),
    url='https://github.com/ovirs/pyvivint',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Home Automation"]
)
